/*
 * Copyright (C) 2024 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if USE(GBM)
#include "DRMDevice.h"
#include <optional>
#include <wtf/Noncopyable.h>
#include <wtf/RefPtr.h>

namespace WebCore {
class GBMDevice;

class DRMDeviceManager {
    WTF_MAKE_NONCOPYABLE(DRMDeviceManager);
    WTF_DEPRECATED_MAKE_FAST_ALLOCATED(DRMDeviceManager);
public:
    static DRMDeviceManager& singleton();

    DRMDeviceManager() = default;
    ~DRMDeviceManager();

    void initializeMainDevice(DRMDevice&&);
    bool isInitialized() const { return m_mainDevice.isInitialized; }
    const DRMDevice& mainDevice() const { return m_mainDevice.device; }

    enum class NodeType : bool { Primary, Render };
    RefPtr<GBMDevice> mainGBMDevice(NodeType) const;
    RefPtr<GBMDevice> gbmDevice(const DRMDevice&, NodeType) const;

private:
    struct {
        bool isInitialized { false };
        DRMDevice device;
        mutable std::optional<RefPtr<GBMDevice>> gbmPrimaryNode;
        mutable std::optional<RefPtr<GBMDevice>> gbmRenderNode;
    } m_mainDevice;
};

} // namespace WebCore

#endif // USE(GBM)
