/*
 * Copyright (C) 2025 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "FilterOperation.h"
#include "StyleColor.h"

namespace WebCore {
namespace Style {

class DropShadowFilterOperationWithStyleColor : public DropShadowFilterOperationBase {
public:
    static Ref<DropShadowFilterOperationWithStyleColor> create(const IntPoint& location, int stdDeviation, const Style::Color& color)
    {
        return adoptRef(*new DropShadowFilterOperationWithStyleColor(location, stdDeviation, color));
    }

    Ref<DropShadowFilterOperation> createEquivalentWithResolvedColor(const RenderStyle&);

    Ref<FilterOperation> clone() const override
    {
        return adoptRef(*new DropShadowFilterOperationWithStyleColor(location(), stdDeviation(), m_styleColor));
    }

    const Style::Color& styleColor() const { return m_styleColor; }

    RefPtr<FilterOperation> blend(const FilterOperation* from, const BlendingContext&, bool blendToPassthrough = false) override;

private:
    bool operator==(const FilterOperation&) const override;
    bool requiresRepaintForCurrentColorChange() const override { return m_styleColor.containsCurrentColor(); }

    void dump(TextStream&) const override;

    DropShadowFilterOperationWithStyleColor(const IntPoint& location, int stdDeviation, const Style::Color& color)
        : DropShadowFilterOperationBase(Type::DropShadowWithStyleColor, location, stdDeviation)
        , m_styleColor(color)
    {
    }

    Style::Color m_styleColor;
};

}
}

SPECIALIZE_TYPE_TRAITS_FILTEROPERATION(Style::DropShadowFilterOperationWithStyleColor, type() == WebCore::FilterOperation::Type::DropShadowWithStyleColor)
