/*
 * Copyright (C) 2022 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <optional>
#include <wtf/WallTime.h>

namespace WebCore::PCM {

enum class AttributionReportEndpoint : bool { Source, Destination };

struct AttributionTimeToSendData {
    std::optional<WallTime> sourceEarliestTimeToSend;
    std::optional<WallTime> destinationEarliestTimeToSend;

    std::optional<WallTime> earliestTimeToSend()
    {
        if (!sourceEarliestTimeToSend && !destinationEarliestTimeToSend)
            return std::nullopt;

        if (sourceEarliestTimeToSend && destinationEarliestTimeToSend)
            return std::min(sourceEarliestTimeToSend, destinationEarliestTimeToSend);

        return sourceEarliestTimeToSend ? sourceEarliestTimeToSend : destinationEarliestTimeToSend;
    }

    std::optional<WallTime> latestTimeToSend()
    {
        if (!sourceEarliestTimeToSend && !destinationEarliestTimeToSend)
            return std::nullopt;

        if (sourceEarliestTimeToSend && destinationEarliestTimeToSend)
            return std::max(sourceEarliestTimeToSend, destinationEarliestTimeToSend);

        return sourceEarliestTimeToSend ? sourceEarliestTimeToSend : destinationEarliestTimeToSend;
    }

    std::optional<AttributionReportEndpoint> attributionReportEndpoint()
    {
        if (sourceEarliestTimeToSend && destinationEarliestTimeToSend) {
            if (*sourceEarliestTimeToSend < *destinationEarliestTimeToSend)
                return AttributionReportEndpoint::Source;

            return AttributionReportEndpoint::Destination;
        }

        if (sourceEarliestTimeToSend)
            return AttributionReportEndpoint::Source;

        if (destinationEarliestTimeToSend)
            return AttributionReportEndpoint::Destination;

        return std::nullopt;
    }
};

}
